unit VocalTract;

interface

uses

  System.SysUtils, System.Math,

  KnobsUtils, KnobsConversions;

//  -------------------------------------------------------------------------------------------------------------------
//
// Original copyright message:
//
//   This file is part of Gnuspeech, an extensible, text-to-speech package, based on real-time, articulatory,
//   speech-synthesis-by-rules.
//
//   Copyright 1991-2012 David R. Hill, Leonard Manzara, Craig Schock
//
//  -------------------------------------------------------------------------------------------------------------------
//
// Translated to Pascal by Blue Hell, copyright note for the changes:
//
//   (C) COPYRIGHT 2019 .. 2020 Blue Hell / Jan Punter
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License version 2 as
//   published by the Free Software Foundation;
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//   For all listed email addresses :
//
//     _dot. to be substituted by a dot      '.'
//     2@t2  to be substituted by an at sign '@'
//
//
//   Blue Hell is a trade mark owned by
//
//     Jan Punter
//     https:// www.bluehell.nl/
//     jan2@t2mail_dot_bluehell_dot_nl
//
//  -------------------------------------------------------------------------------------------------------------------
//
//
//  *******************************************************************************************************************
//
//  REVISION INFORMATION
//  $Author: len $
//  $Date: 1995/04/17 19:51:21 $
//  $Revision: 1.9 $
//  $Source: /cvsroot/softwareTRM/tube.c,v $
//  $State: Exp $
//
//
//  $Log: tube.c,v $
//   * Revision 1.10 2009-04-19 15:14 david
//   * Initial 0.7 release -- added pthread and buffers needed to interface with Objective-C
//   * aComponents created to implement 'Synthesizer' for the Macintosh under OS X
//   * Note that the frication volume final output increases when glottal volume decreases.
//   * This needs to be fixed not
//   *
//   * Revision 1.9 2006/04/01 18:11 david
//   * Remove all things not required for running Synthesizer and split off structures
//   *
//   * Revision 1.8  1995/04/17  19:51:21  len
//   * Temporary fix to frication balance.
//   *
//   * Revision 1.7  1995/03/21  04:52:37  len
//   * Now compiles FAT.  Also adjusted mono and stereo output volume to match
//   * approximately the output volume of the DSP.
//   *
//   * Revision 1.6  1995/03/04  05:55:57  len
//   * Changed controlRate parameter to a Single.
//   *
//   * Revision 1.5  1995/03/02  04:33:04  len
//   * Added amplitude scaling to input of vocal tract and throat, to keep the
//   * software TRM in line with the DSP version.
//   *
//   * Revision 1.4  1994/11/24  05:24:12  len
//   * Added Hi/Low output sample rate case.
//   *
//   * Revision 1.3  1994/10/20  21:20:19  len
//   * Changed nose and mouth aperture filter coefficients, so now specified as
//   * Hz values ( which scale appropriately as the tube length changes), rather
//   * than arbitrary coefficient values ( which don't scale).
//   *
//   * Revision 1.2  1994/08/05  03:12:52  len
//   * Resectioned tube so that it more closely conforms the the DRM proportions.
//   * Also changed frication injection so now allowed from S3 to S10.
//   *
//   * Revision 1.1.1.1  1994/07/07  03:48:52  len
//   * Initial archived version.
//   *
//
//  *******************************************************************************************************************
//
//
//  *******************************************************************************************************************
//  *
//  *     aProgram:      tube
//  *
//  *     Description:   Software ( non-aReal-time) implementation of the Tube
//  *                    Resonance Model for speech production.
//  *
//  *                    note BH: it can prolly be RT nowadays, that's the hope at least.
//  *
//  *     Author:        Leonard Manzara
//  *
//  *     Date:          July 5th, 1994
//  *
//  *******************************************************************************************************************

type

  TVTLogger = procedure( const aSender: TObject; const aMsg: string) of object;

  TVTLoggable = class
  private
    FLogger : TVTLogger;
  private
    function    GetLogger : TVTLogger;                                                                          virtual;
    procedure   SetLogger( const aValue: TVTLogger);                                                            virtual;
  protected
    procedure   Log( const aMsg: string);
    procedure   LogFmt( const aFmt: string; const anArgs: array of const);
  public
    property    Logger: TVTLogger read GetLogger write FLogger;
  end;


  TNoise = class( TVTLoggable)
  private
    FSeed: TSignal;
  public
    constructor Create;
    function    NextValue: TSignal;
  end;


  TNoiseFilter = class( TVTLoggable)
  private
    FNoiseX : TSignal;
  public
    function    NextValue( anInput: TSignal): TSignal;
  end;


  TMouthNasalFilter = class( TVTLoggable)
  private
    FCoeff       : TSignal;
    FA10         : TSignal;
    FA20         : TSignal;
    FA21         : TSignal;
    FB11         : TSignal;
    FB21         : TSignal;
    FReflectionY : TSignal;
    FRadiationX  : TSignal;
    FRadiationY  : TSignal;
  private
    procedure   SetCoeff( aValue: TSignal);
  public
    function    NextReflectionValue( anInput: TSignal): TSignal;
    function    NextRadiationValue ( anInput: TSignal): TSignal;
  public
    property    Coeff: TSignal read FCoeff write SetCoeff;
  end;


  TThroat = class( TVTLoggable)
  private
    FA0         : TSignal;
    FB1         : TSignal;
    FY          : TSignal;
    FSampleRate : TSignal;
    FCutoff     : TSignal;
    FGain       : TSignal;
  private
    procedure   Initialize;
    procedure   SetSampleRate( aValue: TSignal);
    procedure   SetCutoff    ( aValue: TSignal);
  public
    function    NextValue( anInput: TSignal): TSignal;
  public
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
    property    Cutoff     : TSignal read FCutoff     write SetCutoff;
    property    Gain       : TSignal read FGain       write FGain;
  end;


  TBandpassFilter = class( TVTLoggable)
  private
    FXn1             : TSignal;
    FXn2             : TSignal;
    FYn1             : TSignal;
    FYn2             : TSignal;
    FAlpha           : TSignal;
    FBeta            : TSignal;
    FGamma           : TSignal;
    FSampleRate      : TSignal;
    FBandwidth       : TSignal;
    FCenterFrequency : TSignal;
  private
    procedure   Initialize;
    procedure   SetSampleRate     ( aValue: TSignal);
    procedure   SetBandWidth      ( aValue: TSignal);
    procedure   SetCenterFrequency( aValue: TSignal);
  public
    function    NextValue( anInput: TSignal): TSignal;
  public
    property    SampleRate      : TSignal read FSampleRate      write SetSampleRate;
    property    Bandwidth       : TSignal read FBandwidth       write SetBandWidth;
    property    CenterFrequency : TSignal read FCenterFrequency write SetCenterFrequency;
  end;


  TVocalTract = class( TVTLoggable)
  strict private
  const
    TOP                      =  0;
    BOTTOM                   =  1;
    TOTAL_ALPHA_COEFFICIENTS =  3;
    TOTAL_NASAL_SECTIONS     =  6;
    TOTAL_REGIONS            =  8;
    TOTAL_SECTIONS           = 10;
    TOTAL_COEFFICIENTS       = TOTAL_REGIONS;
    TOTAL_NASAL_COEFFICIENTS = TOTAL_NASAL_SECTIONS;

    // OROPHARYNX REGIONS
    R1  = 0;      //  S1
    R2  = 1;      //  S2
    R3  = 2;      //  S3
    R4  = 3;      //  S4 & S5
    R5  = 4;      //  S6 & S7
    R6  = 5;      //  S8
    R7  = 6;      //  S9
    R8  = 7;      //  S10

    // OROPHARYNX SCATTERING JUNCTION COEFFICIENTS (BETWEEN EACH REGION)
    C1  = R1;     //  R1-R2 (S1-S2)
    C2  = R2;     //  R2-R3 (S2-S3)
    C3  = R3;     //  R3-R4 (S3-S4)
    C4  = R4;     //  R4-R5 (S5-S6)
    C5  = R5;     //  R5-R6 (S7-S8)
    C6  = R6;     //  R6-R7 (S8-S9)
    C7  = R7;     //  R7-R8 (S9-S10)
    C8  = R8;     //  R8-AIR (S10-AIR)

    // OROPHARYNX SECTIONS
    S1  = 0;      //  R1
    S2  = 1;      //  R2
    S3  = 2;      //  R3
    S4  = 3;      //  R4
    S5  = 4;      //  R4
    S6  = 5;      //  R5
    S7  = 6;      //  R5
    S8  = 7;      //  R6
    S9  = 8;      //  R7
    S10 = 9;      //  R8

    // FRICATION INJECTION COEFFICIENTS
    FC1                     = 0;      //  S3
    FC2                     = 1;      //  S4
    FC3                     = 2;      //  S5
    FC4                     = 3;      //  S6
    FC5                     = 4;      //  S7
    FC6                     = 5;      //  S8
    FC7                     = 6;      //  S9
    FC8                     = 7;      //  S10
    TOTAL_FRIC_COEFFICIENTS = 8;

    // THREE-WAY JUNCTION ALPHA COEFFICIENTS
    LEFT  = 0;
    RIGHT = 1;
    UPPER = 2;

    // NASAL TRACT SECTIONS
    N1     = 0;
    NVELUM = N1;
    N2     = 1;
    N3     = 2;
    N4     = 3;
    N5     = 4;
    N6     = 5;

    // NASAL TRACT COEFFICIENTS
    NC1 = N1;     //  N1-N2
    NC2 = N2;     //  N2-N3
    NC3 = N3;     //  N3-N4
    NC4 = N4;     //  N4-N5
    NC5 = N5;     //  N5-N6
    NC6 = N6;     //  N6-AIR
  private
    FMouthReflectionFilter : TMouthNasalFilter;
    FNasalReflectionFilter : TMouthNasalFilter;
    FSampleRate            : TSignal;
    FDampingFactor         : TSignal;
    FFrication             : TSignal;
    FCurrPtr               : Integer;
    FPrevPtr               : Integer;
    FApScale               : TSignal;
    FVelum                 : TSignal;
    FFricationAmplitude    : TSignal;
    FFricationPosition     : Double;
    FRadius                : array[ 0 .. TOTAL_REGIONS - 1] of TSignal;
    FOropharynx            : array[ 0 .. TOTAL_SECTIONS -1, 0 .. 1, 0 .. 1] of TSignal;
    FOropharynxCoeff       : array[ 0 .. TOTAL_COEFFICIENTS - 1] of TSignal;
    FFricationTap          : array[ 0 .. TOTAL_FRIC_COEFFICIENTS - 1] of TSignal;
    FAlpha                 : array[ 0 .. TOTAL_ALPHA_COEFFICIENTS - 1] of TSignal;
    FNoseRadius            : array[ 0 .. TOTAL_NASAL_SECTIONS - 1] of TSignal;
    FNasal                 : array[ 0 .. TOTAL_NASAL_SECTIONS - 1, 0 .. 1, 0 .. 1] of TSignal;
    FNasalCoeff            : array[ 0 .. TOTAL_NASAL_COEFFICIENTS - 1  ] of TSignal;
  private
    procedure   SetLogger( const aValue: TVTLogger);                                                           override;
  private
    procedure   SetSampleRate          ( aValue : TSignal);
    function    GetMouthReflectionCoeff         : TSignal;
    procedure   SetMouthReflectionCoeff( aValue : TSignal);
    function    GetNasalReflectionCoeff         : TSignal;
    procedure   SetNasalReflectionCoeff( aValue : TSignal);
    procedure   SetApScale( aValue : TSignal);
    procedure   SetVelum ( aValue : TSignal);
    procedure   SetFricationAmplitude( aValue: TSignal);
    procedure   SetFricationPosition( aValue: TSignal);
    function    GetRadius( anIndex: Integer): TSignal;
    procedure   SetRadius( anIndex: Integer; aValue: TSignal);
    function    GetNoseRadius( anIndex: Integer): TSignal;
    procedure   SetNoseRadius( anIndex: Integer; aValue: TSignal);
  private
    procedure   CalculateTubeCoefficients;
    procedure   SetFricationTaps;
  protected
    procedure   Initialize;
  public
    constructor Create;
    destructor  Destroy;                                                                                       override;
    function    NextValue( anInput, aFrication: TSignal): TSignal;
  public
    property    SampleRate    : TSignal read FSampleRate    write SetSampleRate;
    property    DampingFactor : TSignal read FDampingFactor write FDampingFactor;
    property    Frication     : TSignal read FFrication     write FFrication;
  public
    property    MouthReflectionCoeff          : TSignal read GetMouthReflectionCoeff write SetMouthReflectionCoeff;
    property    NasalReflectionCoeff          : TSignal read GetNasalReflectionCoeff write SetNasalReflectionCoeff;
    property    ApScale                       : TSignal read FApScale                write SetApScale;
    property    Velum                         : TSignal read FVelum                  write SetVelum;
    property    FricationAmplitude            : TSignal read FFricationAmplitude     write SetFricationAmplitude;
    property    FricationPosition             : TSignal read FFricationPosition      write SetFricationPosition;
    property    Radius    [ anIndex: Integer] : TSignal read GetRadius               write SetRadius;
    property    NoseRadius[ anIndex: Integer] : TSignal read GetNoseRadius           write SetNoseRadius;
  end;


  TSynthesizer = class( TVTLoggable)
  private
    FNoise          : TNoise;
    FNoiseFilter    : TNoiseFilter;
    FBandPassFilter : TBandPassFilter;
    FVocalTract     : TVocalTract;
    FThroat         : TThroat;
  private
    FSampleRate     : TSignal;
    FGlotPitch      : TSignal;
    FFrequency      : TSignal;
    FGlotVolume     : TSignal;
    FAspVolume      : TSignal;
    FBreathiness    : TSignal;
    FCrossMix       : TSignal;
    FModulation     : Boolean;
  private
    procedure   SetLogger( const aValue: TVTLogger);                                                           override;
  private
    procedure   SetSampleRate                   ( aValue : TSignal);
    procedure   SetGlotPitch                    ( aValue : TSignal);
    function    GetApScale                               : TSignal;
    procedure   SetApScale                      ( aValue : TSignal);
    function    GetVelum                                 : TSignal;
    procedure   SetVelum                        ( aValue : TSignal);
    function    GetFricationAmplitude                    : TSignal;
    procedure   SetFricationAmplitude           ( aValue : TSignal);
    function    GetFricationPosition                     : TSignal;
    procedure   SetFricationPosition            ( aValue : TSignal);
    function    GetRadius    ( anIndex : Integer)        : TSignal;
    procedure   SetRadius    ( anIndex : Integer; aValue : TSignal);
    function    GetNoseRadius( anIndex : Integer)        : TSignal;
    procedure   SetNoseRadius( anIndex : Integer; aValue : TSignal);
    function    GetBandpassBandwidth                     : TSignal;
    procedure   SetBandpassBandwidth            ( aValue : TSignal);
    function    GetBandpassCenterFrequency               : TSignal;
    procedure   SetBandpassCenterFrequency      ( aValue : TSignal);
    function    GetThroatCutoff                          : TSignal;
    procedure   SetThroatCutoff                 ( aValue : TSignal);
    function    GetThroatGain                            : TSignal;
    procedure   SetThroatGain                   ( aValue : TSignal);
    function    GetMouthReflectionCoeff                  : TSignal;
    procedure   SetMouthReflectionCoeff         ( aValue : TSignal);
    function    GetNasalReflectionCoeff                  : TSignal;
    procedure   SetNasalReflectionCoeff         ( aValue : TSignal);
  private
    procedure   SetDefaults;
  public
    constructor Create;
    procedure   Initialize;
    destructor  Destroy;                                                                                       override;
    function    NextValue( anInput: TSignal): TSignal;
  public
    property    SampleRate                    : TSignal read FSampleRate                write SetSampleRate;
    property    GlotPitch                     : TSignal read FGlotPitch                 write SetGlotPitch;
    property    GlotVolume                    : TSignal read FGlotVolume                write FGlotVolume;
    property    AspVolume                     : TSignal read FAspVolume                 write FAspVolume;
    property    Breathiness                   : TSignal read FBreathiness               write FBreathiness;
    property    CrossMix                      : TSignal read FCrossMix                  write FCrossMix;
    property    Modulation                    : Boolean read FModulation                write FModulation;
    property    BandPassBandwidth             : TSignal read GetBandpassBandwidth       write SetBandpassBandwidth;
    property    BandpassCenterFrequency       : TSignal read GetBandpassCenterFrequency write SetBandpassCenterFrequency;
    property    ThroatCutoff                  : TSignal read GetThroatCutoff            write SetThroatCutoff;
    property    ThroatGain                    : TSignal read GetThroatGain              write SetThroatGain;
    property    MouthReflectionCoeff          : TSignal read GetMouthReflectionCoeff    write SetMouthReflectionCoeff;
    property    NasalReflectionCoeff          : TSignal read GetNasalReflectionCoeff    write SetNasalReflectionCoeff;
    property    ApScale                       : TSignal read GetApScale                 write SetApScale;
    property    Velum                         : TSignal read GetVelum                   write SetVelum;
    property    FricationAmplitude            : TSignal read GetFricationAmplitude      write SetFricationAmplitude;
    property    FricationPosition             : TSignal read GetFricationPosition       write SetFricationPosition;
    property    Radius    [ anIndex: Integer] : TSignal read GetRadius                  write SetRadius;
    property    NoseRadius[ anIndex: Integer] : TSignal read GetNoseRadius              write SetNoseRadius;
  end;



implementation


{ ========
  TVTLoggable = class
  private
    FLogger : TVTLogger;
  public
    property    Logger: TVTLogger read GetLogger write FLogger;
  private
}

    function    TVTLoggable.GetLogger : TVTLogger; // virtual;
    begin
      Result:= FLogger
    end;


    procedure   TVTLoggable.SetLogger( const aValue: TVTLogger); // virtual;
    begin
      FLogger := aValue;
    end;


//  protected

    procedure   TVTLoggable.Log( const aMsg: string);
    begin
      if Assigned( FLogger)
      then FLogger( Self, aMsg);
    end;


    procedure   TVTLoggable.LogFmt( const aFmt: string; const anArgs: array of const);
    begin
      Log( Format( aFmt, anArgs, AppLocale));
    end;


{ ========
  TNoise = class( TVTLoggable)
  private
    FSeed: TSignal;
  public
}
    constructor TNoise.Create;
    begin
      FSeed := 0.7892347;
    end;


    function    TNoise.NextValue: TSignal;
    var
      aProduct : TSignal;
    begin
      aProduct := 377.0 * FSeed;
      FSeed    := aProduct - Floor( aProduct);
      Result   := FSeed - 0.5;
    end;


{ ========
  TNoiseFilter = class( TVTLoggable)
  private
    FNoiseX : TSignal;
  public
}

    function    TNoiseFilter.NextValue( anInput: TSignal): TSignal;
    begin
      Result  := anInput + FNoiseX;
      FNoiseX := anInput;
    end;


{ ========
  TMouthNasalFilter = class( TVTLoggable)
  private
    FCoeff       : TSignal;
    FA10         : TSignal;
    FA20         : TSignal;
    FA21         : TSignal;
    FB11         : TSignal;
    FB21         : TSignal;
    FReflectionY : TSignal;
    FRadiationX  : TSignal;
    FRadiationY  : TSignal;
  public
    property    Coeff: TSignal read FCoeff write SetCoeff;
  private
}

    procedure   TMouthNasalFilter.SetCoeff( aValue: TSignal);
    begin
      if aValue <> FCoeff
      then begin
        FCoeff := aValue;
        FB11   := - aValue;
        FA10   :=   1.0 - Abs( FB11);
        FA20   :=   aValue;
        FB21   := - FA20;
        FA21   :=   FB21;
      end;
    end;


//  public

    function    TMouthNasalFilter.NextReflectionValue( anInput: TSignal): TSignal;
    begin
      Result       := FA10 * anInput + FB11 * FReflectionY;
      FReflectionY := Result;
    end;


    function    TMouthNasalFilter.NextRadiationValue ( anInput: TSignal): TSignal;
    begin
      Result      := FA20 * anInput + FA21 * FRadiationX - FB21 * FRadiationY;
      FRadiationX := anInput;
      FRadiationY := Result;
    end;


{ ========
  TThroat = class( TVTLoggable)
  private
    FA0         : TSignal;
    FB1         : TSignal;
    FY          : TSignal;
    FSampleRate : TSignal;
    FCutoff     : TSignal;
    FGain       : TSignal;
  public
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
    property    Cutoff     : TSignal read FCutoff     write SetCutoff;
    property    Gain       : TSignal read FGain       write FGain;
  private
}

    procedure   TThroat.Initialize;
    begin
      FA0 := 2.0 * FCutoff / FSampleRate;
      FB1 := 1.0 - FA0;
    end;


    procedure   TThroat.SetSampleRate( aValue: TSignal);
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate := aValue;
        Initialize;
      end;
    end;


    procedure   TThroat.SetCutoff( aValue: TSignal);
    begin
      if aValue <> FCutoff
      then begin
        FCutoff := aValue;
        Initialize;
      end;
    end;


//  public

    function    TThroat.NextValue( anInput: TSignal): TSignal;
    begin
      Result := FA0 * anInput + FB1 * FY;
      FY     := Result;
      Result := FGain * Result;
    end;


{ ========
  TBandpassFilter = class( TVTLoggable)
  private
    FXn1             : TSignal;
    FXn2             : TSignal;
    FYn1             : TSignal;
    FYn2             : TSignal;
    FAlpha           : TSignal;
    FBeta            : TSignal;
    FGamma           : TSignal;
    FSampleRate      : TSignal;
    FBandwidth       : TSignal;
    FCenterFrequency : TSignal;
  public
    property    SampleRate      : TSignal write SetSampleRate;
    property    BandWidth       : TSignal write SetBandWidth;
    property    CenterFrequency : TSignal write SetCenterFrequency;
  private
}

    procedure   TBandpassFilter.Initialize;
    var
      aTan : TSignal;
      aCos : TSignal;
    begin
      aTan := Tan( PI * FBandWidth / FSampleRate);
      aCos := Cos( 2.0 * PI * FCenterFrequency / FSampleRate);

      FBeta  := ( 1.0 - aTan) / ( 2.0 * ( 1.0 + aTan));
      FGamma := ( 0.5 + FBeta) * aCos;
      FAlpha := ( 0.5 - FBeta) / 2.0;
    end;


    procedure   TBandpassFilter.SetSampleRate( aValue: TSignal);
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate := aValue;
        Initialize;
      end;
    end;


    procedure   TBandpassFilter.SetBandWidth( aValue: TSignal);
    begin
      if aValue <> FBandWidth
      then begin
        FBandWidth := aValue;
        Initialize;
      end;
    end;


    procedure   TBandpassFilter.SetCenterFrequency( aValue: TSignal);
    begin
      if aValue <> FCenterFrequency
      then begin
        FCenterFrequency := aValue;
        Initialize;
      end;
    end;


//  public

    function    TBandpassFilter.NextValue( anInput: TSignal): TSignal;
    begin
      Result := 2.0 * (( FAlpha * ( anInput - FXn2)) + ( FGamma * FYn1) - ( FBeta * FYn2));
      FXn2   := FXn1;
      FXn1   := anInput;
      FYn2   := FYn1;
      FYn1   := Result;
    end;


{ ========
  TVocalTract = class( TVTLoggable)
  private
    FMouthReflectionFilter : TMouthNasalFilter;
    FNasalReflectionFilter : TMouthNasalFilter;
    FSampleRate            : TSignal;
    FDampingFactor         : TSignal;
    FFrication             : TSignal;
    FCurrPtr               : Integer;
    FPrevPtr               : Integer;
    FRadius                : array[ 0 .. TOTAL_REGIONS - 1] of TSignal;
    FOropharynx            : array[ 0 .. TOTAL_SECTIONS -1, 0 .. 1, 0 .. 1] of TSignal;
    FOropharynxCoeff       : array[ 0 .. TOTAL_COEFFICIENTS - 1] of TSignal;
    FFricationTap          : array[ 0 .. TOTAL_FRIC_COEFFICIENTS - 1] of TSignal;
    FAlpha                 : array[ 0 .. TOTAL_ALPHA_COEFFICIENTS - 1] of TSignal;
    FNasal                 : array[ 0 .. TOTAL_NASAL_SECTIONS - 1, 0 .. 1, 0 .. 1] of TSignal;
    FNasalCoeff            : array[ 0 .. TOTAL_NASAL_COEFFICIENTS - 1  ] of TSignal;
  public
    property    SampleRate    : TSignal read FSampleRate    write SetSampleRate;
    property    DampingFactor : TSignal read FDampingFactor write FDampingFactor;
    property    Frication     : TSignal read FFrication     write FFrication;
  public
    property    MouthReflectionCoeff      : TSignal read GetMouthReflectionCoeff write SetMouthReflectionCoeff;
    property    NasalReflectionCoeff      : TSignal read GetNasalReflectionCoeff write SetNasalReflectionCoeff;
    property    Radius[ anIndex: Integer] : TSignal read GetRadius               write SetRadius;
  private
}

    procedure   TVocalTract.SetLogger( const aValue: TVTLogger); // override;
    begin
      inherited;
      FMouthReflectionFilter.Logger := aValue;
      FNasalReflectionFilter.Logger := aValue;
    end;


//  private

    procedure   TVocalTract.SetSampleRate( aValue: TSignal);
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate := aValue;
      end;
    end;


    function    TVocalTract.GetMouthReflectionCoeff: TSignal;
    begin
      Result := FMouthReflectionFilter.Coeff;
    end;


    procedure   TVocalTract.SetMouthReflectionCoeff( aValue : TSignal);
    begin
      FMouthReflectionFilter.Coeff := aValue;
    end;


    function    TVocalTract.GetNasalReflectionCoeff: TSignal;
    begin
      Result := FNasalReflectionFilter.Coeff;
    end;


    procedure   TVocalTract.SetNasalReflectionCoeff( aValue : TSignal);
    begin
      FNasalReflectionFilter.Coeff := aValue;
    end;


    procedure   TVocalTract.SetApScale( aValue : TSignal);
    begin
      if aValue <> FApScale
      then begin
        FApScale := aValue;
        CalculateTubeCoefficients;
      end;
    end;


    procedure   TVocalTract.SetVelum( aValue : TSignal);
    begin
      if aValue <> FVelum
      then begin
        FVelum := aValue;
        CalculateTubeCoefficients;
      end;
    end;


    procedure   TVocalTract.SetFricationAmplitude( aValue: TSignal);
    begin
      if aValue <> FFricationAmplitude
      then begin
        FFricationAmplitude := aValue;
        SetFricationTaps;
      end;
    end;


    procedure   TVocalTract.SetFricationPosition( aValue: TSignal);
    begin
      if aValue <> FFricationPosition
      then begin
        FFricationPosition := aValue;
        SetFricationTaps;
      end;
    end;


    function    TVocalTract.GetRadius( anIndex: Integer): TSignal;
    begin
      if   ( anIndex >= Low ( FRadius))
      and  ( anIndex <= High( FRadius))
      then Result := FRadius[ anIndex]
      else Result := 0.0;
    end;


    procedure   TVocalTract.SetRadius( anIndex: Integer; aValue: TSignal);
    begin
      if   ( anIndex >= Low ( FRadius))
      and  ( anIndex <= High( FRadius))
      then begin
        FRadius[ anIndex] := aValue;
        CalculateTubeCoefficients;
      end;
    end;


    function    TVocalTract.GetNoseRadius( anIndex: Integer): TSignal;
    begin
      if   ( anIndex >= Low ( FNoseRadius))
      and  ( anIndex <= High( FNoseRadius))
      then Result := FNoseRadius[ anIndex]
      else Result := 0.0;
    end;


    procedure   TVocalTract.SetNoseRadius( anIndex: Integer; aValue: TSignal);
    begin
      if   ( anIndex >= Low ( FNoseRadius))
      and  ( anIndex <= High( FNoseRadius))
      then begin
        FNoseRadius[ anIndex] := aValue;
        CalculateTubeCoefficients;
      end;
    end;


//  private

    procedure   TVocalTract.CalculateTubeCoefficients;
    var
      i     : Integer;
      radA2 : TSignal;
      radB2 : TSignal;
      r0_2  : TSignal;
      r1_2  : TSignal;
      r2_2  : TSignal;
      sum   : TSignal;
    begin
      //  CALCULATE COEFFICIENTS FOR THE OROPHARYNX
      for i := 0 to TOTAL_REGIONS - 2
      do begin
        radA2                := Radius[ i    ] * Radius[ i    ];
        radB2                := Radius[ i + 1] * Radius[ i + 1];
        FOropharynxCoeff[ i] := ( radA2 - radB2) / ( radA2 + radB2);
      end;

      //  CALCULATE THE COEFFICIENT FOR THE MOUTH APERTURE
      radA2 := Radius[ R8] * Radius[ R8];
      radB2 := ApScale     * ApScale;
      FOropharynxCoeff[ C8] := ( radA2 - radB2) / ( radA2 + radB2);

      //  CALCULATE ALPHA COEFFICIENTS FOR 3-WAY JUNCTION
      //  NOTE:  SINCE JUNCTION IS IN MIDDLE OF REGION 4, r0_2 = r1_2
      r1_2 := Radius[ R4] * Radius[R4];
      r0_2 := r1_2;
      r2_2 := Velum * Velum;
      sum  := 2.0 / ( r0_2 + r1_2 + r2_2);
      FAlpha[ LEFT ] := sum * r0_2;
      FAlpha[ RIGHT] := sum * r1_2;
      FAlpha[ UPPER] := sum * r2_2;

      //  AND 1ST NASAL PASSAGE COEFFICIENT
      radA2 := FVelum * FVelum;
    //printf("current.velum is %f", current.velum);
      radB2 := FNoseRadius[ N2] * FNoseRadius[ N2];
      FNasalCoeff[ NC1] := ( radA2 - radB2) / ( radA2 + radB2);
    end;


    procedure   TVocalTract.SetFricationTaps;
    var
      i            : Integer;
      IntegerPart  : Integer;
      Complement   : TSignal;
      Remainder    : TSignal;
      FricationAmp : Double;
    begin
      FricationAmp := FricationAmplitude * 10.0; // Volume x 10 to be audible. Investigate
      IntegerPart  := Floor( FricationPosition);
      Complement   := FricationPosition - IntegerPart;
      Remainder    := 1.0 - Complement;
      i            := FC1;

      while i < TOTAL_FRIC_COEFFICIENTS
      do begin
        FFricationTap[ i] := Remainder * FricationAmp;

        if   i < TOTAL_FRIC_COEFFICIENTS - 1
        then begin
          Inc( i);
          FFricationTap[ i] := Complement * FricationAmp;
        end
        else FFricationTap[ i] := 0.0;

        Inc( i);
      end;
    end;


//  protected

    procedure   TVocalTract.Initialize;
    begin
      CalculateTubeCoefficients;
      SetFricationTaps;
    end;


//  public

    constructor TVocalTract.Create;
    begin
      FMouthReflectionFilter := TMouthNasalFilter.Create;
      FNasalReflectionFilter := TMouthNasalFilter.Create;
    end;


    destructor  TVocalTract.Destroy; // override;
    begin
      FreeAndNil( FNasalReflectionFilter);
      FreeAndNil( FMouthReflectionFilter);
    end;


    function    TVocalTract.NextValue( anInput, aFrication: TSignal): TSignal;
    var
      i                : Integer;
      j                : Integer;
      k                : Integer;
      Delta            : TSignal;
      JunctionPressure : TSignal;
    begin
      FCurrPtr := ( FCurrPtr + 1) mod 2;
      FPrevPtr := ( FPrevPtr + 1) mod 2;

      FOropharynx[ S1][ TOP][ FCurrPtr] := ( FOropharynx[ S1][ BOTTOM][ FPrevPtr] * dampingFactor) + anInput;

      // CALCULATE THE SCATTERING JUNCTIONS FOR S1-S2
      Delta := FOropharynxCoeff[ C1] * ( FOropharynx[ S1][ TOP][ FPrevPtr] - FOropharynx[ S2][ BOTTOM][ FPrevPtr]);
      FOropharynx[ S2][ TOP][ FCurrPtr] := ( FOropharynx[ S1][ TOP][ FPrevPtr] + Delta) * dampingFactor;
      FOropharynx[ S1][ BOTTOM][ FCurrPtr] := ( FOropharynx[ S2][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;

      // CALCULATE THE SCATTERING JUNCTIONS FOR S2-S3 AND S3-S4

      j := C2;
      k := FC1;

      for i := S2 to S4 - 1
      do begin
        Delta := FOropharynxCoeff[ j] * ( FOropharynx[ i][ TOP][ FPrevPtr] - FOropharynx[ i + 1][ BOTTOM][ FPrevPtr]);
        FOropharynx[ i + 1][ TOP][ FCurrPtr] := (( FOropharynx[ i][ TOP][ FPrevPtr] + Delta) * dampingFactor) + ( FFricationTap[ k] * frication);
        FOropharynx[ i][ BOTTOM][ FCurrPtr] := ( FOropharynx[ i + 1][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;
        Inc( j);
        Inc( k);
       end;

      // UPDATE 3-WAY JUNCTION BETWEEN THE MIDDLE OF R4 AND NASAL CAVITY
      JunctionPressure := ( FAlpha[ LEFT] * FOropharynx[ S4][ TOP][ FPrevPtr]) + ( FAlpha[ RIGHT] * FOropharynx[ S5][ BOTTOM][ FPrevPtr]) + ( FAlpha[ UPPER] * FNasal[ NVELUM][ BOTTOM][ FPrevPtr]);
      FOropharynx[ S4][ BOTTOM][ FCurrPtr] := ( JunctionPressure - FOropharynx[ S4][ TOP][ FPrevPtr]) * dampingFactor;
      FOropharynx[ S5][ TOP][ FCurrPtr] := (( JunctionPressure - FOropharynx[ S5][ BOTTOM][ FPrevPtr]) * dampingFactor) + ( FFricationTap[ FC3] * frication);
      FNasal[ NVELUM][ TOP][ FCurrPtr] := ( JunctionPressure - FNasal[ NVELUM][ BOTTOM][ FPrevPtr]) * dampingFactor;

      // CALCULATE JUNCTION BETWEEN R4 AND R5 ( S5-S6)
      Delta := FOropharynxCoeff[ C4] * ( FOropharynx[ S5][ TOP][ FPrevPtr] - FOropharynx[ S6][ BOTTOM][ FPrevPtr]);
      FOropharynx[ S6][ TOP][ FCurrPtr] := (( FOropharynx[ S5][ TOP][ FPrevPtr] + Delta) * dampingFactor) + ( FFricationTap[ FC4] * frication);
      FOropharynx[ S5][ BOTTOM][ FCurrPtr] := ( FOropharynx[ S6][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;

      // CALCULATE JUNCTION INSIDE R5 ( S6-S7) ( PURE DELAY WITH DAMPING)
      FOropharynx[ S7][ TOP][ FCurrPtr] := ( FOropharynx[ S6][ TOP][ FPrevPtr] * dampingFactor) + ( FFricationTap[ FC5] * frication);
      FOropharynx[ S6][ BOTTOM][ FCurrPtr] := FOropharynx[ S7][ BOTTOM][ FPrevPtr] * dampingFactor;

      // CALCULATE LAST 3 INTERNAL JUNCTIONS ( S7-S8, S8-S9, S9-S10)

      j := C5;
      k := FC6;

      for i := S7 to S10 - 1
      do begin
        Delta := FOropharynxCoeff[ j] * ( FOropharynx[ i][ TOP][ FPrevPtr] - FOropharynx[ i + 1][ BOTTOM][ FPrevPtr]);
        FOropharynx[ i + 1][ TOP][ FCurrPtr] := (( FOropharynx[ i][ TOP][ FPrevPtr] + Delta) * dampingFactor) + ( FFricationTap[ k] * frication);
        FOropharynx[ i][ BOTTOM][ FCurrPtr] := ( FOropharynx[ i + 1][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;
        Inc( j);
        Inc( k);
      end;

      // REFLECTED SIGNAL AT MOUTH GOES THROUGH A LOWPASS FILTER
      FOropharynx[ S10][ BOTTOM][ FCurrPtr] :=  dampingFactor * FMouthReflectionFilter.NextReflectionValue( FOropharynxCoeff[ C8] * FOropharynx[ S10][ TOP][ FPrevPtr]);

      // OUTPUT FROM MOUTH GOES THROUGH A HIGHPASS FILTER
      Result := FMouthReflectionFilter.NextRadiationValue(( 1.0 + FOropharynxCoeff[ C8]) * FOropharynx[ S10][ TOP][ FPrevPtr]);

      // UPDATE NASAL CAVITY
      j := NC1;

      for i := NVELUM to N6 - 1
      do begin
        Delta := FNasalCoeff[ j] * ( FNasal[ i][ TOP][ FPrevPtr] - FNasal[ i + 1][ BOTTOM][ FPrevPtr]);
        FNasal[ i + 1][ TOP][ FCurrPtr] := ( FNasal[ i][ TOP][ FPrevPtr] + Delta) * dampingFactor;
        FNasal[ i][ BOTTOM][ FCurrPtr] := ( FNasal[ i + 1][ BOTTOM][ FPrevPtr] + Delta) * dampingFactor;
        Inc( j);
      end;

      // REFLECTED SIGNAL AT NOSE GOES THROUGH A LOWPASS FILTER
      FNasal[ N6][ BOTTOM][ FCurrPtr] := dampingFactor * FNasalReflectionFilter.NextReflectionValue( FNasalCoeff[ NC6] * FNasal[ N6][ TOP][ FPrevPtr]);

      // OUTPUT FROM NOSE GOES THROUGH A HIGHPASS FILTER
      Result := Result + FNasalReflectionFilter.NextRadiationValue(( 1.0 + FNasalCoeff[ NC6]) * FNasal[ N6][ TOP][ FPrevPtr]);
    end;


{ ========
  TSynthesizer = class( TVTLoggable)
  private
    FNoise          : TNoise;
    FNoiseFilter    : TNoiseFilter;
    FBandPassFilter : TBandPassFilter;
    FVocalTract     : TVocalTract;
    FThroat         : TThroat;
  private
    FSampleRate     : TSignal;
    FGlotPitch      : TSignal;
    FFrequency      : TSignal;
    FGlotVolume     : TSignal;
    FAspVolume      : TSignal;
    FBreathiness    : TSignal;
    FCrossMix       : TSignal;
    FModulation     : Boolean;
  public
    property    SampleRate  : TSignal  read FSampleRate  write SetSampleRate;
    property    GlotPitch   : TSignal  read FGlotPitch   write SetGlotPitch;
    property    GlotVolume  : TSignal  read FGlotVolume  write FGlotVolume;
    property    AspVolume   : TSignal  read FAspVolume   write FAspVolume;
    property    Breathiness : TSignal  read FBreathiness write FBreathiness;
    property    CrossMix    : TSignal  read FCrossMix    write FCrossMix;
    property    Modulation  : Boolean read FModulation  write FModulation;
  public
    property    BandPassBandwidth       : TSignal read GetBandpassBandwidth       write SetBandpassBandwidth;
    property    BandpassCenterFrequency : TSignal read GetBandpassCenterFrequency write SetBandpassCenterFrequency;
    property    ThroatCutoff            : TSignal read GetThroatCutoff            write SetThroatCutoff;
    property    ThroatGain              : TSignal read GetThroatGain              write SetThroatGain;
    property    MouthReflectionCoeff    : TSignal read GetMouthReflectionCoeff    write SetMouthReflectionCoeff;
    property    NasalReflectionCoeff    : TSignal read GetNasalReflectionCoeff    write SetNasalReflectionCoeff;
  private
}

    procedure   TSynthesizer.SetLogger( const aValue: TVTLogger); // override;
    begin
      inherited;
      FNoise         .Logger := aValue;
      FNoiseFilter   .Logger := aValue;
      FBandPassFilter.Logger := aValue;
      FVocalTract    .Logger := aValue;
      FThroat        .Logger := aValue;
    end;


//  private

    procedure   TSynthesizer.SetSampleRate( aValue : TSignal);
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate                := aValue;
        FBandpassFilter.SampleRate := aValue;
        FVocalTract    .SampleRate := aValue;
        FThroat        .SampleRate := aValue;
      end;
    end;


    procedure   TSynthesizer.SetGlotPitch( aValue : TSignal);
    begin
      if aValue <> FGlotPitch
      then begin
        FGlotPitch := aValue;
        FFrequency := UnitsToFrequency( aValue);
      end;
    end;


    function    TSynthesizer.GetApScale: TSignal;
    begin
      Result := FVocalTract.ApScale;
    end;


    procedure   TSynthesizer.SetApScale( aValue : TSignal);
    begin
      FVocalTract.ApScale := aValue;
    end;


    function    TSynthesizer.GetVelum: TSignal;
    begin
      Result := FVocalTract.Velum;
    end;


    procedure   TSynthesizer.SetVelum( aValue : TSignal);
    begin
      FVocalTract.Velum := aValue;
    end;


    function    TSynthesizer.GetFricationAmplitude: TSignal;
    begin
      Result := FVocalTract.FricationAmplitude;
    end;


    procedure   TSynthesizer.SetFricationAmplitude( aValue : TSignal);
    begin
      FVocalTract.FricationAmplitude := aValue;
    end;


    function    TSynthesizer.GetFricationPosition: TSignal;
    begin
      Result := FVocalTract.FricationPosition;
    end;


    procedure   TSynthesizer.SetFricationPosition( aValue : TSignal);
    begin
      FVocalTract.FricationPosition := aValue;;
    end;


    function    TSynthesizer.GetRadius( anIndex : Integer): TSignal;
    begin
      Result := FVocalTract.Radius[ anIndex];
    end;


    procedure   TSynthesizer.SetRadius( anIndex : Integer; aValue: TSignal);
    begin
      FVocalTract.Radius[ anIndex] := aValue;
    end;


    function    TSynthesizer.GetNoseRadius( anIndex : Integer): TSignal;
    begin
      Result := FVocalTract.NoseRadius[ anIndex];
    end;


    procedure   TSynthesizer.SetNoseRadius( anIndex : Integer; aValue: TSignal);
    begin
      FVocalTract.NoseRadius[ anIndex] := aValue;
    end;


//  private

    function    TSynthesizer.GetBandpassBandwidth: TSignal;
    begin
      Result := FBandPassFilter.Bandwidth;
    end;


    procedure   TSynthesizer.SetBandpassBandwidth( aValue : TSignal);
    begin
      FBandPassFilter.Bandwidth := aValue;
    end;


    function    TSynthesizer.GetBandpassCenterFrequency: TSignal;
    begin
      Result := FBandPassFilter.CenterFrequency;
    end;


    procedure   TSynthesizer.SetBandpassCenterFrequency( aValue : TSignal);
    begin
      FBandPassFilter.CenterFrequency := aValue;
    end;


    function    TSynthesizer.GetThroatCutoff: TSignal;
    begin
      Result := FThroat.Cutoff;
    end;


    procedure   TSynthesizer.SetThroatCutoff( aValue : TSignal);
    begin
      FThroat.Cutoff := aValue;
    end;


    function    TSynthesizer.GetThroatGain: TSignal;
    begin
      Result := FThroat.Gain;
    end;


    procedure   TSynthesizer.SetThroatGain( aValue : TSignal);
    begin
      FThroat.Gain := aValue;
    end;


    function    TSynthesizer.GetMouthReflectionCoeff: TSignal;
    begin
      Result := FVocalTract.MouthReflectionCoeff;
    end;


    procedure   TSynthesizer.SetMouthReflectionCoeff( aValue : TSignal);
    begin
      FVocalTract.MouthReflectionCoeff := aValue;
    end;


    function    TSynthesizer.GetNasalReflectionCoeff: TSignal;
    begin
      Result := FVocalTract.NasalReflectionCoeff;
    end;


    procedure   TSynthesizer.SetNasalReflectionCoeff( aValue : TSignal);
    begin
      FVocalTract.NasalReflectionCoeff := aValue;
    end;


//  private

    procedure   TSynthesizer.SetDefaults;
    begin
      GlotPitch               :=  -12.0; // 220 Hz
      GlotVolume              :=    1.0; // Max volume
      AspVolume               :=    0.5;
      Breathiness             :=    2.5;
      BandpassCenterFrequency := 5000.0;
      BandpassBandWidth       :=  250.0;
      CrossMix                :=    0.5;
      Modulation              := True;
      ThroatCutoff            := 1600.0;
      ThroatGain              :=    0.5;
      MouthReflectionCoeff    := 4000.0;
      NasalReflectionCoeff    := 4000.0;
      ApScale                 :=    2.5;
      Velum                   :=    0.0;
      FricationAmplitude      :=    0.0;
      FricationPosition       :=    8.0;
      Radius[ 0]              :=    0.8;
      Radius[ 1]              :=    1.67;
      Radius[ 2]              :=    1.905;
      Radius[ 3]              :=    1.985;
      Radius[ 4]              :=    0.81;
      Radius[ 5]              :=    0.495;
      Radius[ 6]              :=    0.73;
      Radius[ 7]              :=    1.485;
      NoseRadius[ 0]          :=    1.35;
      NoseRadius[ 1]          :=    1.35;
      NoseRadius[ 2]          :=    1.7;
      NoseRadius[ 3]          :=    1.7;
      NoseRadius[ 4]          :=    1.3;
      NoseRadius[ 5]          :=    0.9;
    end;


//  public

    constructor TSynthesizer.Create;
    begin
      FNoise          := TNoise         .Create;
      FNoiseFilter    := TNoiseFilter   .Create;
      FBandPassFilter := TBandPassFilter.Create;
      FVocalTract     := TVocalTract    .Create;
      FThroat         := TThroat        .Create;
      SetDefaults;
    end;


    procedure   TSynthesizer.Initialize;
    begin
    end;


    destructor  TSynthesizer.Destroy; // override;
    begin
      FreeAndNil( FThroat        );
      FreeAndNil( FVocalTract    );
      FreeAndNil( FBandPassFilter);
      FreeAndNil( FNoiseFilter   );
      FreeAndNil( FNoise         );
    end;


    function    TSynthesizer.NextValue( anInput: TSignal): TSignal;
    var
      aNoise       : TSignal;
      aPulsedNoise : TSignal;
      aPulse       : TSignal;
      aCrossMix    : TSignal;
    begin
      aPulse       := anInput;
      aNoise       := FNoiseFilter.NextValue( FNoise.NextValue);
      aPulsedNoise := aNoise * aPulse;
      aPulse       := GlotVolume * ( aPulse * ( 1.0 - Breathiness) + aPulsedNoise * Breathiness);

      if   Modulation
      then begin
        aCrossMix := Min( GlotVolume * CrossMix, 1.0);
        Result    := aPulsedNoise * aCrossMix + ( 1.0 - aCrossMix) * aNoise;
      end
      else Result := aNoise;

      Result := FVocalTract.NextValue(( aPulse + AspVolume * Result) * 0.125 {VT_SCALE}, FBandpassFilter.NextValue( Result));
      Result := Result + FThroat.NextValue( aPulse * 0.125 {VT_SCALE});
      Result := Result * 100;
    end;


end.

