unit poetry_generator;

interface

uses

  KnobsUtils;

type

  TPoetryGenerator = class
  //
  // Source  : http://www.jonathanmackenzie.net/portfolio/PHD.pdf
  // Details : USING STRANGE ATTRACTORS. TO MODEL SOUND. Submitted to. The University of
  //           London for the Degree of Doctor of Philosophy. Jonathan Mackenzie.
  // Also    : http://www.jonathanmackenzie.net/portfolio/project_phd.html
  // Uses    : The Poetry Generation Algorithm from Chapter 8 of the Thesis.
  //
  private
    type TState = record
      FHistory     : TSignalArray;
      FSeed        : TSignalArray;
      FInPointer   : Integer;
      FSeedPointer : Integer;
    end;
    PState = ^ TState;
  private
    FPolyphony : Integer;
    FEpsilon   : TSignal;
    FState     : array of TState;
  private
    function    GetHistoryCount( aVoice: Integer) : Integer;
    function    GetSeedCount   ( aVoice: Integer) : Integer;
  private
    function    Equal( A, B: TSignal): Boolean;
    function    RandomFindSeed( aVoice: Integer): Integer;
    procedure   ShiftSeed     ( aVoice: Integer; aValue: TSignal);
  private
    property    Polyphony                      : Integer read FPolyphony;
    property    HistoryCount[ aVoice: Integer] : Integer read GetHistoryCount;
    property    SeedCount   [ aVoice: Integer] : Integer read GetSeedCount;
  public
    constructor Create( aPolyphony: Integer; aSeedLength, aHistorySize: Word; anEpsilon: TSignal);
    procedure   AddHistory    ( aVoice: Integer; const aValue: TSignal);
    procedure   MakeRandomSeed( aVoice: Integer);
    function    Tick          ( aVoice: Integer): TSignal;
  end;



implementation


uses

 Math;


{ ========
  TPoetryGenerator = class
  private
    type TState = record
      FHistory     : TSignalArray;
      FSeed        : TSignalArray;
      FInPointer   : Integer;
      FSeedPointer : Integer;
    end;
    PState = ^ TState;
  private
    FPolyphony : Integer;
    FEpsilon   : TSignal;
    FState     : array of TState;
  public
    property    Polyphony                      : Integer read FPolyphony;
    property    HistoryCount[ aVoice: Integer] : Integer read GetHistoryCount;
    property    SeedCount   [ aVoice: Integer] : Integer read GetSeedCount;
  private
}

    function    TPoetryGenerator.GetHistoryCount( aVoice: Integer): Integer;
    begin
      Result := Length( FState[ aVoice].FHistory);
    end;


    function    TPoetryGenerator.GetSeedCount( aVoice: Integer): Integer;
    begin
      Result := Length( FState[ aVoice].FSeed);
    end;


//  private

    function    TPoetryGenerator.Equal( A, B: TSignal): Boolean;
    begin
      Result := Abs( A - B) < FEpsilon;
    end;


    function    TPoetryGenerator.RandomFindSeed( aVoice: Integer): Integer;
    var
      State : PState;
      p     : Integer;
      i     : Integer;
      j     : Integer;
      Found : Boolean;
    begin
      State  := @ FState[ aVoice];
      Result := -1;
      p      := Random( HistoryCount[ aVoice]);                  // Start Seed search at a random position
      i      := 0;

      while i < HistoryCount[ aVoice]
      do begin
        Found := True;
        j     := 0;

        while j < SeedCount[ aVoice]
        do begin
          if Equal(
            State^.FHistory[ ( p + i + j) mod HistoryCount[ aVoice]],
            State^.FSeed[ ( State^.FSeedPointer + j) mod SeedCount[ aVoice]]
          )
          then Inc( j)
          else begin
            Found := False;
            Inc( i);
            Break;
          end
        end;

        if Found
        then begin
          Result := ( i + p) mod HistoryCount[ aVoice];
          Break;
        end;

      end;
    end;


    procedure   TPoetryGenerator.ShiftSeed( aVoice: Integer; aValue: TSignal);
    var
      State : PState;
    begin
      State  := @ FState[ aVoice];
      State^.FSeedPointer := ( State^.FSeedPointer + 1) mod SeedCount[ aVoice];
      State^.FSeed[ ( State^.FSeedPointer + SeedCount[ aVoice] - 1) mod SeedCount[ aVoice]] := aValue;
    end;

//  public


    constructor TPoetryGenerator.Create( aPolyphony: Integer; aSeedLength, aHistorySize: Word; anEpsilon: TSignal);
    var
      zz    : Integer;
      State : PState;
      i     : Integer;
    begin
      inherited Create;
      FPolyphony := aPolyphony;
      SetLength( FState, Polyphony);
      FEpsilon := anEpsilon;

      for zz := 0 to Polyphony - 1
      do begin
        State := @ FState[ zz];
        SetLength( State^.FHistory, aHistorySize);
        SetLength( State^.FSeed   , aSeedLength );

        for i := 0 to SeedCount[ zz] - 1
        do State^.FSeed[ i] := 0;

        for i := 0 to HistoryCount[ zz] - 1
        do State^.FHistory[ i] := 0;

        State^.FInPointer   := 0;
        State^.FSeedPointer := 0;
      end;
    end;


    procedure   TPoetryGenerator.AddHistory( aVoice: Integer; const aValue: TSignal);
    var
      State : PState;
    begin
      State := @ FState[ aVoice];
      State^.FHistory[ State^.FInPointer] := aValue;
      State^.FInPointer := ( State^.FInPointer + 1) mod HistoryCount[ aVoice];
      MakeRandomSeed( aVoice);
    end;


    procedure   TPoetryGenerator.MakeRandomSeed( aVoice: Integer);
    var
      State : PState;
      p     : Integer;
      i     : Integer;
    begin
      State := @ FState[ aVoice];
      p := Random( HistoryCount[ aVoice]);

      for i := 0 to SeedCount[ aVoice] - 1
      do begin
        State^.FSeed[ ( State^.FSeedPointer + i) mod SeedCount[ aVoice]] := State^.FHistory[ p];
        p := ( p + 1) mod HistoryCount[ aVoice];
      end;
    end;


    function    TPoetryGenerator.Tick( aVoice: Integer): TSignal;
    var
      State : PState;
      i     : Integer;
    begin
      State := @ FState[ aVoice];
      i := RandomFindSeed( aVoice);

      if i >= 0
      then ShiftSeed( aVoice, State^.FHistory[ ( i + SeedCount[ aVoice]) mod HistoryCount[ aVoice]])
      else MakeRandomSeed( aVoice);

      Result := State^.FSeed[ State^.FSeedPointer];
    end;


end.

