unit walshy;

{$ifdef DEBUG}
//  {$define walshy_test}
{$endif}

{
  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

  source: https://github.com/bvssvni/fwht

  Original release notes:

  FWHT - Fast Walsh Hadamard Transform in C

    BSD license.
    by Sven Nilsen, 2012
    http://www.cutoutpro.com

    Version: 0.000 in angular degrees version notation
    http://isprogrammingeasy.blogspot.no/2012/08/angular-degrees-versioning-notation.html

  This implementation was made without looking at other FWHT algorithms,
  so I release under its own license.

  -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

  Translated to Delphi for use in Wren.

  For modifications only:

  (C) COPYRIGHT modifications 2019 .. 2020 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://www.bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}

interface

uses

  KnobsUtils;


  procedure WalshTransform       ( const aSrc, aDst, aTmp1, aTmp2: TSignalArray);
  procedure WalshNormalize       ( const aSrc: TSignalArray);
  function  WalshSumAbsDifference( const a, b: TSignalArray): TSignal;


implementation


  procedure WalshTransform( const aSrc, aDst, aTmp1, aTmp2: TSignalArray);
  var
    a     : PSignalArray;
    b     : PSignalArray;
    tmp   : PSignalArray;
    n     : Integer;
    i     : Integer;
    j     : Integer;
    s     : Integer;
    v1    : Integer;
    v2    : Integer;
    v3    : Integer;
  begin
    n := Length( aSrc);
    Assert( n = Length( aDst));
    Assert( n > 0);
    Assert( IsPowerOfTwo( n));

    Move( aSrc[ 0], aTmp1[ 0], n * SizeOf( aSrc[ 0]));
    a := @ aTmp1;
    b := @ aTmp2;
    i := n shr 1;

    while i > 0
    do begin
      for j := 0 to n -1
      do begin
        s := ( j div i) mod 2;

        if   s <> 0
        then begin v1 := -i; v2 := -1; v3 :=  0; end
        else begin v1 :=  0; v2 :=  1; v3 :=  i; end;

        b^[ j] := a^[ v1 + j] + v2 * a^[ v3 + j];
      end;

      tmp := a;
      a   := b;
      b   := tmp;
      i   := i shr 1;
    end;

    Move( a^[ 0], aDst[ 0], n * SizeOf( aSrc[ 0]));
  end;


  procedure WalshNormalize( const aSrc: TSignalArray);
  var
    i : Integer;
    n : Integer;
  begin
    n := Length( aSrc);

    for i := 0 to n - 1
    do aSrc[ i] := aSrc[ i] / n;
  end;


  function  WalshSumAbsDifference( const a, b: TSignalArray): TSignal;
  var
    Sum  : TSignal;
    Diff : TSignal;
    i    : Integer;
  begin
    Assert( Length( a) = Length( b));
    Assert( Length( a) > 0);
    Sum := 0;

    for i := 0 to Length( a) - 1
    do begin
      Diff := b[ i] - a[ i];
      Sum  := Sum + Abs( Diff);
    end;

    Result := Sum / Length( a);
  end;


{$ifdef walshy_test}

  // ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  // Unit test stuff below

  procedure Err( Where: Integer);
  begin
    Nop;
  end;


  procedure TestStuff;
  var
    i       : Integer;
    t1      : TSignalArray;
    t2      : TSignalArray;
    a       : TSignalArray;
    b       : TSignalArray;
    a2      : TSignalArray;
    b2      : TSignalArray;
    answ    : TSignalArray;
    diff_a  : TSignal;
    diff_b  : TSignal;
    diff_a2 : TSignal;
    diff_b2 : TSignal;
    diff_a3 : TSignal;
    diff_b3 : TSignal;
  begin
    SetLength( a , 8);
    SetLength( b , 8);
    SetLength( t1, 8);
    SetLength( t2, 8);

    a[ 0] := 1;
    a[ 1] := 0;
    a[ 2] := 1;
    a[ 3] := 0;
    a[ 4] := 0;
    a[ 5] := 1;
    a[ 6] := 1;
    a[ 7] := 0;

    WalshTransform( a, b, t1, t2);

    if b[ 0] <>  4 then Err(  0);
    if b[ 1] <>  2 then Err(  1);
    if b[ 2] <>  0 then Err(  2);
    if b[ 3] <> -2 then Err(  3);
    if b[ 4] <>  0 then Err(  4);
    if b[ 5] <>  2 then Err(  5);
    if b[ 6] <>  0 then Err(  6);
    if b[ 7] <>  2 then Err(  7);

    Nop;

    WalshTransform( b, a, t1, t2);
    WalshNormalize( a);

    if a[ 0] <>  1 then Err(  8);
    if a[ 1] <>  0 then Err(  9);
    if a[ 2] <>  1 then Err( 10);
    if a[ 3] <>  0 then Err( 11);
    if a[ 4] <>  0 then Err( 12);
    if a[ 5] <>  1 then Err( 13);
    if a[ 6] <>  1 then Err( 14);
    if a[ 7] <>  0 then Err( 15);

    Nop;

    SetLength( a   , 16);
    SetLength( b   , 16);
    SetLength( t1  , 16);
    SetLength( t2  , 16);
    SetLength( answ, 16);

    a[  0] := 1; answ[  0] := a[  0];
    a[  1] := 0; answ[  1] := a[  1];
    a[  2] := 1; answ[  2] := a[  2];
    a[  3] := 0; answ[  3] := a[  3];
    a[  4] := 0; answ[  4] := a[  4];
    a[  5] := 1; answ[  5] := a[  5];
    a[  6] := 1; answ[  6] := a[  6];
    a[  7] := 0; answ[  7] := a[  7];
    a[  8] := 1; answ[  8] := a[  8];
    a[  9] := 1; answ[  9] := a[  9];
    a[ 10] := 1; answ[ 10] := a[ 10];
    a[ 11] := 0; answ[ 11] := a[ 11];
    a[ 12] := 1; answ[ 12] := a[ 12];
    a[ 13] := 0; answ[ 13] := a[ 13];
    a[ 14] := 0; answ[ 14] := a[ 14];
    a[ 15] := 0; answ[ 15] := a[ 15];


    WalshTransform( a, b, t1, t2);
    WalshTransform( b, a, t1, t2);
    WalshNormalize( a);

    Nop;

    for i := 0 to Length( a) - 1
    do begin
      if   a[ i] <> answ[ i]
      then Err( 100 + i);
    end;

    Nop;

    SetLength( a , 8);
    SetLength( b , 8);
    SetLength( t1, 8);
    SetLength( t2, 8);

    a[ 0] := 1; b[ 0] := 1;
    a[ 1] := 0; b[ 1] := 0;
    a[ 2] := 1; b[ 2] := 0;
    a[ 3] := 0; b[ 3] := 0;
    a[ 4] := 0; b[ 4] := 0;
    a[ 5] := 1; b[ 5] := 1;
    a[ 6] := 1; b[ 6] := 1;
    a[ 7] := 0; b[ 7] := 0;

    diff_a := WalshSumAbsDifference( a, b);
    diff_b := WalshSumAbsDifference( b, a);

    if   diff_a <> diff_b
    then Err( 200);

    SetLength( a2, Length( a));
    SetLength( b2, Length( b));
    SetLength( t1, Length( a));
    SetLength( t2, Length( b));
    WalshTransform( a, a2, t1, t2);
    WalshTransform( b, b2, t1, t2);

    diff_a2 := WalshSumAbsDifference( a2, b2);
    diff_b2 := WalshSumAbsDifference( b2, a2);

    if   diff_a2 <> diff_b2
    then Err( 201);

    WalshTransform( a2, a, t1, t2);
    WalshTransform( b2, b, t1, t2);
    WalshNormalize( a);
    WalshNormalize( b);

    diff_a3 := WalshSumAbsDifference( a, b);
    diff_b3 := WalshSumAbsDifference( b, a);

    if   diff_a <> diff_a3
    then Err( 202);

    if   diff_b <> diff_b3
    then Err( 203);

    Nop;
  end;



initialization

  TestStuff;

{$endif}

end.

